"use strict";
/*
 * adonis-responsive-attachment
 *
 * (c) Ndianabasi Udonkang <ndianabasi@furnish.ng>
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ResponsiveAttachment = void 0;
/// <reference path="../../adonis-typings/index.ts" />
const utils_1 = require("@poppinss/utils");
const ImageManipulationHelper_1 = require("../Helpers/ImageManipulationHelper");
const lodash_1 = __importDefault(require("lodash"));
const cuid_1 = __importDefault(require("cuid"));
const decorator_1 = require("./decorator");
const REQUIRED_ATTRIBUTES = [
    'name',
    'size',
    'extname',
    'mimeType',
    'width',
    'height',
    'hash',
    'breakpoints',
    'format',
];
/**
 * Attachment class represents an attachment data type
 * for Lucid models
 */
class ResponsiveAttachment {
    constructor(attributes) {
        /**
         * Find if the image has been persisted or not.
         */
        this.isPersisted = false;
        /**
         * "isLocal = true" means the instance is created locally
         * using the bodyparser file object
         */
        this.isLocal = !!this.path;
        this.name = attributes.name;
        this.size = attributes.size;
        this.hash = attributes.hash;
        this.width = attributes.width;
        this.format = attributes.format;
        this.height = attributes.height;
        this.extname = attributes.extname;
        this.mimeType = attributes.mimeType;
        this.url = attributes.url || null;
        this.breakpoints = attributes.breakpoints || undefined;
        this.path = attributes.path || '';
        this.isLocal = !!this.path;
        this.urls = null;
    }
    /**
     * Reference to the drive
     */
    static getDrive() {
        return this.drive;
    }
    /**
     * Set the drive instance
     */
    static setDrive(drive) {
        this.drive = drive;
    }
    /**
     * Create attachment instance from the bodyparser
     * file
     */
    static async fromFile(file) {
        if (file) {
            // Store the file locally first and add the path to the ImageInfo
            // This will be removed after the operation is completed
            await file.moveToDisk('image_upload_tmp');
            const attributes = {
                extname: file.extname,
                mimeType: `${file.type}/${file.subtype}`,
                size: file.size,
                path: file.filePath,
            };
            const responsiveAttachment = new ResponsiveAttachment(attributes);
            return responsiveAttachment;
        }
        else
            return null;
    }
    /**
     * Create attachment instance from the database response
     */
    static fromDbResponse(response) {
        const attributes = typeof response === 'string' ? JSON.parse(response) : response;
        if (!attributes)
            return null;
        /**
         * Validate the db response
         */
        REQUIRED_ATTRIBUTES.forEach((attribute) => {
            if (attributes[attribute] === undefined) {
                throw new utils_1.Exception(`Cannot create attachment from database response. Missing attribute "${attribute}"`);
            }
        });
        const attachment = new ResponsiveAttachment(attributes);
        /**
         * Images fetched from DB are always persisted
         */
        attachment.isPersisted = true;
        return attachment;
    }
    get attributes() {
        return {
            name: this.name,
            size: this.size,
            hash: this.hash,
            width: this.width,
            format: this.format,
            height: this.height,
            extname: this.extname,
            mimeType: this.mimeType,
            url: this.url,
            breakpoints: this.breakpoints,
            path: this.path,
        };
    }
    /**
     * Returns disk instance
     */
    getDisk() {
        const disk = this.options?.disk;
        const Drive = this.constructor.getDrive();
        return disk ? Drive.use(disk) : Drive.use();
    }
    /**
     * Define persistance options
     */
    setOptions(options) {
        this.options = lodash_1.default.merge({
            preComputeUrls: false,
            breakpoints: decorator_1.DEFAULT_BREAKPOINTS,
            forceFormat: undefined,
            optimizeOrientation: true,
            optimizeSize: true,
            responsiveDimensions: true,
        }, options);
        return this;
    }
    async enhanceFile() {
        // Read the image as a buffer using `Drive.get()`
        const originalFileBuffer = await this.getDisk().get(this.path);
        // Optimise the image buffer and return the optimised buffer
        // and the info of the image
        const { buffer, info } = await (0, ImageManipulationHelper_1.optimize)(originalFileBuffer, this.options);
        // Override the `imageInfo` object with the optimised `info` object
        // As the optimised `info` object is preferred
        // Also append the `hash` and `buffer`
        return lodash_1.default.assign({ ...this.attributes }, info, { hash: (0, cuid_1.default)(), buffer });
    }
    /**
     * Save image to the disk. Results in noop when "this.isLocal = false"
     */
    async save() {
        /**
         * Do not persist already persisted image or if the
         * instance is not local
         */
        if (!this.isLocal || this.isPersisted) {
            return;
        }
        /**
         * Read the original temporary file from disk and optimise the file while
         * return the enhanced buffer and information of the enhanced buffer
         */
        const enhancedImageData = await this.enhanceFile();
        /**
         * Generate the name of the original image
         */
        this.name = (0, ImageManipulationHelper_1.generateName)({
            extname: enhancedImageData.extname,
            hash: enhancedImageData.hash,
            options: this.options,
            prefix: 'original',
        });
        /**
         * Update the local attributes with the attributes
         * of the optimised original file
         */
        this.size = enhancedImageData.size;
        this.hash = enhancedImageData.hash;
        this.width = enhancedImageData.width;
        this.format = enhancedImageData.format;
        this.height = enhancedImageData.height;
        this.extname = enhancedImageData.extname;
        this.mimeType = enhancedImageData.mimeType;
        /**
         * Inject the name into the `ImageInfo`
         */
        enhancedImageData.name = this.name;
        /**
         * Write the optimised original image to the disk
         */
        await this.getDisk().put(enhancedImageData.name, enhancedImageData.buffer);
        //await this.file!.moveToDisk('./', { name: this.name }, this.options?.disk)
        /**
         * Generate image thumbnail data
         */
        const thumbnailImageData = await (0, ImageManipulationHelper_1.generateThumbnail)(enhancedImageData, this.options);
        if (thumbnailImageData) {
            /**
             * Write the thumbnail image to the disk
             */
            await this.getDisk().put(thumbnailImageData.name, thumbnailImageData.buffer);
            /**
             * Delete buffer from `thumbnailImageData`
             */
            delete thumbnailImageData.buffer;
            lodash_1.default.set(enhancedImageData, 'breakpoints.thumbnail', thumbnailImageData);
        }
        /**
         * Generate breakpoint image data
         */
        const breakpointFormats = await (0, ImageManipulationHelper_1.generateBreakpointImages)(enhancedImageData, this.options);
        if (breakpointFormats && Array.isArray(breakpointFormats) && breakpointFormats.length > 0) {
            for (const format of breakpointFormats) {
                if (!format)
                    continue;
                const { key, file: breakpointImageData } = format;
                /**
                 * Write the breakpoint image to the disk
                 */
                await this.getDisk().put(breakpointImageData.name, breakpointImageData.buffer);
                /**
                 * Delete buffer from `breakpointImageData`
                 */
                delete breakpointImageData.buffer;
                lodash_1.default.set(enhancedImageData, ['breakpoints', key], breakpointImageData);
            }
        }
        const { width, height } = await (0, ImageManipulationHelper_1.getDimensions)(enhancedImageData.buffer);
        delete enhancedImageData.buffer;
        delete enhancedImageData.path;
        lodash_1.default.assign(enhancedImageData, {
            width,
            height,
        });
        /**
         * Update the local value of `breakpoints`
         */
        this.breakpoints = enhancedImageData.breakpoints;
        /**
         * Images has been persisted
         */
        this.isPersisted = true;
        /**
         * Forcefully compute the URL
         */
        await this.computeUrls();
        /**
         * Delete the temporary file
         */
        await this.getDisk().delete(this.path);
        return lodash_1.default.merge(enhancedImageData, this.urls);
    }
    /**
     * Delete original and responsive images from the disk
     */
    async delete() {
        if (!this.isPersisted) {
            return;
        }
        /**
         * Delete the original image
         */
        await this.getDisk().delete(this.name);
        /**
         * Delete the responsive images
         */
        if (this.breakpoints) {
            for (const key in this.breakpoints) {
                if (Object.prototype.hasOwnProperty.call(this.breakpoints, key)) {
                    const breakpointImage = this.breakpoints[key];
                    await this.getDisk().delete(breakpointImage.name);
                }
            }
        }
        this.isDeleted = true;
        this.isPersisted = false;
    }
    async computeUrls(options) {
        const { forced, signedUrlOptions } = options || {};
        /**
         * Cannot compute url for a non persisted image
         */
        if (!this.isPersisted) {
            return;
        }
        /**
         * Compute urls when preComputeUrls is set to true
         * or the `preComputeUrls` function exists
         * or the computation is forced
         */
        if (!this.options?.preComputeUrls && !forced) {
            return;
        }
        const disk = this.getDisk();
        /**
         * Generate url using the user defined preComputeUrls method
         */
        if (typeof this.options?.preComputeUrls === 'function') {
            const urls = await this.options.preComputeUrls(disk, this);
            this.url = urls.url;
            if (!this.urls)
                this.urls = {};
            if (!this.urls.breakpoints)
                this.urls.breakpoints = {};
            for (const key in urls.breakpoints) {
                if (Object.prototype.hasOwnProperty.call(urls.breakpoints, key)) {
                    if (!this.urls.breakpoints[key])
                        this.urls.breakpoints[key] = { url: '' };
                    this.urls.breakpoints[key].url = urls.breakpoints[key].url;
                }
            }
            return;
        }
        /**
         * Iterative URL-computation logic
         */
        const { path, ...originalAttributes } = this.attributes;
        const attachmentData = originalAttributes;
        if (attachmentData) {
            if (!this.urls)
                this.urls = {};
            for (const key in attachmentData) {
                if (['name', 'breakpoints'].includes(key) === false)
                    continue;
                const value = attachmentData[key];
                let url;
                if (key === 'name') {
                    const name = value;
                    const imageVisibility = await disk.getVisibility(name);
                    if (imageVisibility === 'private') {
                        url = await disk.getSignedUrl(name, signedUrlOptions || undefined);
                    }
                    else {
                        url = await disk.getUrl(name);
                    }
                    this.urls['url'] = url;
                    this.url = url;
                }
                else if (key === 'breakpoints') {
                    if (lodash_1.default.isEmpty(value) !== true) {
                        if (!this.urls.breakpoints)
                            this.urls.breakpoints = {};
                        for (const breakpoint in value) {
                            if (Object.prototype.hasOwnProperty.call(value, breakpoint)) {
                                const breakpointImageData = value?.[breakpoint];
                                if (breakpointImageData) {
                                    const imageVisibility = await disk.getVisibility(breakpointImageData.name);
                                    if (imageVisibility === 'private') {
                                        url = await disk.getSignedUrl(breakpointImageData.name, signedUrlOptions || undefined);
                                    }
                                    else {
                                        url = await disk.getUrl(breakpointImageData.name);
                                    }
                                    this.urls['breakpoints'][breakpoint] = { url };
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    /**
     * Returns the signed or unsigned URL for each responsive image
     */
    async getUrls(signingOptions) {
        /**
         * Forcefully compute the URLs first
         */
        await this.computeUrls({ forced: true, ...signingOptions });
        return this.urls;
    }
    /**
     * Serialize attachment instance to JSON
     */
    toJSON() {
        const { path, ...originalAttributes } = this.attributes;
        return {
            ...originalAttributes,
            breakpoints: this.breakpoints,
        };
    }
}
exports.ResponsiveAttachment = ResponsiveAttachment;
